r"""
Copy with a memory.

(c) Tom de Geus, 2021, MIT
"""
import argparse
import os
import sys

import numpy as np
import prettytable

from . import checksum
from . import convert
from . import detail
from . import path
from . import relpath
from . import rich
from . import rsync
from . import scp
from . import ssh
from . import yaml
from ._version import version
from ._version import version_tuple
from .cli.defaults import f_dump
from .cli.defaults import f_hostinfo


def _shelephant_diff_parser():
    """
    Return parser for :py:func:`shelephant_diff`.
    """

    desc = "Compare local and remote files and list differences."
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument("--version", action="version", version=version)
    parser.add_argument("-f", "--force", action="store_true", help="Force overwrite output.")
    parser.add_argument("--yaml", type=str, help="Dump as YAML file.")
    parser.add_argument("--sort", type=str, help="Sort printed table by column.")
    parser.add_argument("--table", type=str, default="SINGLE_BORDER", help="Select print style.")
    parser.add_argument("--checksum", action="store_true", help="Check sums.")
    parser.add_argument("--get-new", type=str, help="Save to get '<-' files.")
    parser.add_argument("--get-diff", type=str, help="Save to get ''!=' files.")
    parser.add_argument("--get-all", type=str, help="Save to get '<-' and '!=' files.")
    parser.add_argument("--send-new", type=str, nargs=2, help="Save to send '<-' files.")
    parser.add_argument("--send-diff", type=str, nargs=2, help="Save to send ''!=' files.")
    parser.add_argument("--send-all", type=str, nargs=2, help="Save to send '<-' and '!=' files.")
    parser.add_argument(
        "local", type=str, nargs="?", default=f_dump, help="Local files, see shelephant_dump."
    )
    parser.add_argument(
        "remote",
        type=str,
        nargs="?",
        default=f_hostinfo,
        help="Remote files, see shelephant_hostinfo.",
    )
    return parser


def shelephant_diff(args: list[str]):
    """
    Command-line tool to print datasets from a file, see ``--help``.
    :param args: Command-line arguments (should be all strings).
    """

    parser = _shelephant_diff_parser()
    args = parser.parse_args(args)

    local = yaml.read(args.local)
    remote = yaml.read(args.remote)

    if type(local) is list:
        local = {"files": local, "prefix": os.path.dirname(args.local), "list": True}
    if type(remote) is list:
        remote = {"files": remote, "prefix": os.path.dirname(args.remote), "list": True}

    for field in [local, remote]:
        if "host" in field:
            field["dirname"] = field["host"] + ":" + field["prefix"]
        else:
            field["dirname"] = field["prefix"]

        if len(field["dirname"]) == 0:
            field["dirname"] = "."

        field["files"] = np.array(field["files"])

    ret = {}

    to_remote = rsync.diff(
        source_dir=local["dirname"],
        dest_dir=remote["dirname"],
        files=local["files"],
        checksum=args.checksum,
    )
    ret["=="] = list(local["files"][to_remote["skip"]])
    ret["!="] = list(local["files"][to_remote["overwrite"]])
    ret["->"] = list(local["files"][to_remote["create"]])

    from_remote = rsync.diff(
        source_dir=remote["dirname"],
        dest_dir=local["dirname"],
        files=remote["files"],
        checksum=args.checksum,
    )
    ret["=="] += list(remote["files"][from_remote["skip"]])
    ret["!="] += list(remote["files"][from_remote["overwrite"]])
    ret["<-"] = list(remote["files"][from_remote["create"]])

    for key in ["==", "!="]:
        ret[key] = [str(i) for i in np.unique(ret[key])]
    for key in ["->", "<-"]:
        ret[key] = [str(i) for i in ret[key]]

    stop = False

    for filename, value in zip(
        [args.get_new, args.get_diff, args.get_all], [ret["<-"], ret["!="], ret["<-"] + ret["!="]]
    ):
        if filename is not None:
            stop = True
            tmp = {"files": value}
            for key in ["host", "prefix"]:
                if key in remote:
                    tmp[key] = remote[key]
            yaml.dump(filename, tmp, force=args.force)

    for filename, value in zip(
        [args.send_new, args.send_diff, args.send_all],
        [ret["->"], ret["!="], ret["->"] + ret["!="]],
    ):
        if filename is not None:
            assert "host" not in local, "Not supported by shelephant_send."
            assert len(os.path.dirname(args.local)) == 0, "Not supported by shelephant_send."
            stop = True

            yaml.dump(filename[0], value, force=args.force)

            tmp = {}
            for key in ["host", "prefix"]:
                if key in remote:
                    tmp[key] = remote[key]
            yaml.dump(filename[1], tmp, force=args.force)

    if args.yaml is not None:
        yaml.dump(args.yaml, ret, force=args.force)
        stop = True

    if stop:
        return 0

    out = prettytable.PrettyTable()
    if args.table == "PLAIN_COLUMNS":
        out.set_style(prettytable.PLAIN_COLUMNS)
    elif args.table == "SINGLE_BORDER":
        out.set_style(prettytable.SINGLE_BORDER)
    out.field_names = ["local", "sync", "remote"]
    out.align["local"] = "l"
    out.align["sync"] = "c"
    out.align["remote"] = "l"

    for item in ret["!="]:
        out.add_row([item, "!=", item])

    for item in ret["->"]:
        out.add_row([item, "->", ""])

    for item in ret["<-"]:
        out.add_row(["", "<-", item])

    if args.sort is None:
        print(out.get_string())
    else:
        print(out.get_string(sortby=args.sort))


def _shelephant_diff_catch():
    shelephant_diff(sys.argv[1:])



"""Remove files listed in a (field of a) YAML-file.
The filenames are assumed either absolute, or relative to the input YAML-file.

:usage:

    shelephant_rm [options]

    shelephant_rm [options] <input.yaml>

:argument:

    YAML-file with filenames. Default: shelephant_dump.yaml

:options:

    -k, --key=arg
        Path in the YAML-file, separated by "/". [default: /]

    -f, --force
        Remove without prompt.

    -h, --help
        Show help.

    --version
        Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/shelephant
"""
import argparse
import os

import click

from .. import relpath
from .. import version
from .. import yaml
from .defaults import f_dump


def main_impl():
    class Parser(argparse.ArgumentParser):
        def print_help(self):
            print(__doc__)

    parser = Parser()
    parser.add_argument("-k", "--key", default="/")
    parser.add_argument("-f", "--force", action="store_true")
    parser.add_argument("-v", "--version", action="version", version=version)
    parser.add_argument("input", nargs="?", default=f_dump)
    args = parser.parse_args()

    source = args.input
    key = list(filter(None, args.key.split("/")))
    files = yaml.read_item(source, key)
    prefix = os.path.dirname(source)
    files = relpath.add_prefix(prefix, files)

    if len(files) == 0:
        return 0

    for file in files:
        print(f"rm {file:s}")

    if not args.force:
        if not click.confirm("Proceed?"):
            return 1

    for file in files:
        os.remove(file)


def main():
    main_impl()


if __name__ == "__main__":
    main()



import os
import shutil
import subprocess


def run(cmd):
    return subprocess.check_output(cmd, shell=True).decode("utf-8")


for dirname in ["myssh_send", "myssh_get"]:
    if os.path.isdir(dirname):
        shutil.rmtree(dirname)

    os.mkdir(dirname)

with open("myssh_send/bar.txt", "w") as file:
    file.write("bar")

with open("myssh_send/foo.txt", "w") as file:
    file.write("foo")

with open("myssh_get/foo.txt", "w") as file:
    file.write("foo")

run("shelephant_dump -o myssh_send/shelephant_dump.yaml myssh_send/bar.txt myssh_send/foo.txt")
run("shelephant_dump -o myssh_get/shelephant_dump.yaml myssh_get/foo.txt")
run("shelephant_checksum -o myssh_send/shelephant_checksum.yaml myssh_send/shelephant_dump.yaml")
run("shelephant_checksum -o myssh_get/shelephant_checksum.yaml myssh_get/shelephant_dump.yaml")



"""Parse a YAML-file, and print to screen.

:usage:

    shelephant_parse <file.yaml>

:argument:

    File path.

:options:

    -h, --help
        Show help.

    --version
        Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/shelephant
"""
import argparse

from .. import version
from .. import yaml


def main_impl():
    class Parser(argparse.ArgumentParser):
        def print_help(self):
            print(__doc__)

    parser = Parser()
    parser.add_argument("-v", "--version", action="version", version=version)
    parser.add_argument("file")
    args = parser.parse_args()

    data = yaml.read(args.file)
    yaml.view(data)


def main():
    main_impl()


if __name__ == "__main__":
    main()



"""Copy files listed in a (field of a) YAML-file.
The filenames are assumed either absolute, or relative to the input YAML-file.

By default, *rsync* is used to check if files are different.
*rsync* uses basic criteria such as file size and creation and modification date,
see `rsync manual <https://www.samba.org/ftp/rsync/rsync.html>`_.
This is fast but is only approximate.
In addition you can use:

*   ``--checksum``: Have *rsync* compare checksums of files.
*   ``--check-manual``: Do not use *rsync*.
    If ``--checksum`` is used, checksums are compared (that are optionally precomputed).

:usage:

    shelephant_cp [options] <destination>

    shelephant_cp [options] <input.yaml> <destination>

:arguments:

    <input.yaml>
        YAML-file with filenames. Default: shelephant_dump.yaml

    <destination>
        Prefix of the destination.

:options:

    -c, --checksum
        Use checksum to skip files that are the same.

    -M, --check-manual
        Use internal algorithms instead of *rsync*.

    -k, --key=arg
        Path in the YAML-file, separated by "/". [default: /]

    --colors=arg
        Select color scheme from: none, dark. [default: dark]

    -s, --summary
        Print summary (and no details unless specified).

    -d, --details
        Print details (and no summary unless specified).

    -f, --force
        Move without prompt.

    -q, --quiet
        Do not print progress.

    -h, --help
        Show help.

    --version
        Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/shelephant
"""
import argparse
import os
import shutil
import warnings

from .. import detail
from .. import version
from .. import yaml


def main_impl():
    class Parser(argparse.ArgumentParser):
        def print_help(self):
            print(__doc__)

    parser = Parser()
    parser.add_argument("-c", "--checksum", action="store_true")
    parser.add_argument("-M", "--check-manual", action="store_true")
    parser.add_argument("-k", "--key", default="/")
    parser.add_argument("--colors", default="dark")
    parser.add_argument("-s", "--summary", action="store_true")
    parser.add_argument("-d", "--details", action="store_true")
    parser.add_argument("-f", "--force", action="store_true")
    parser.add_argument("-q", "--quiet", action="store_true")
    parser.add_argument("-v", "--version", action="version", version=version)
    parser.add_argument("args", nargs="+")
    args = parser.parse_args()

    use_rsync = True
    if args.check_manual:
        use_rsync = False
    if not shutil.which("rsync"):
        warnings.warn("rsync not found, using internal fallback")
        use_rsync = False

    if len(args.args) == 1:
        source = "shelephant_dump.yaml"
        dest_dir = args.args[0]
    elif len(args.args) == 2:
        source = args.args[0]
        dest_dir = args.args[1]
    else:
        raise OSError("Too many arguments specified")

    key = list(filter(None, args.key.split("/")))

    return detail.copy(
        copy_function=shutil.copy2,
        files=yaml.read_item(source, key),
        src_dir=os.path.dirname(source),
        dest_dir=dest_dir,
        checksum=args.checksum,
        check_rsync=use_rsync,
        quiet=args.quiet,
        force=args.force,
        print_details=not (args.force or args.summary) or args.details,
        print_summary=not (args.force or args.details) or args.summary,
        print_all=args.details,
        theme_name=args.colors.lower(),
    )


def main():
    main_impl()


if __name__ == "__main__":
    main()



r"""
Yaml IO.

(c) Tom de Geus, 2021, MIT
"""
import os

import click
import yaml

from . import convert


def read(filename):
    r"""
    Read YAML file and return its content.
    """

    if not os.path.isfile(filename):
        raise OSError(f'"{filename} does not exist')

    with open(filename) as file:
        return yaml.load(file.read(), Loader=yaml.FullLoader)


def read_item(filename, key=[]):
    r"""
    Get an item from a YAML file.

    :type key: str or list
    :param key:
        The item to read. E.g.
        *   ``[]`` for a YAML file containing only a list.
        *   ``['foo']`` for a plain YAML file.
        *   ``['key', 'to', foo']`` for a YAML file with nested items.

        An item specified as ``str`` separated by "/" is also accepted.

    :return:
        The read item.
    """

    data = read(filename)

    if type(data) == dict:
        return convert.get(data, key)

    key = convert.split_key(key)

    if type(data) == list and len(key) == 0:
        return data

    raise OSError(f'"{"/".join(key)}" not in "{filename}"')


def dump(filename, data, force=False):
    r"""
    Dump data to YAML file.

    :type filename: str
    :param filename: The output filename.

    :type data: list, dict
    :param data: The data to dump.

    :type force: bool, optional
    :param force: Do not prompt to overwrite file.
    """

    dirname = os.path.dirname(filename)

    if not force:
        if os.path.isfile(filename):
            if not click.confirm(f'Overwrite "{filename}"?'):
                raise OSError("Cancelled")
        elif not os.path.isdir(dirname) and len(dirname) > 0:
            if not click.confirm(f'Create "{os.path.dirname(filename)}"?'):
                raise OSError("Cancelled")

    if not os.path.isdir(dirname) and len(dirname) > 0:
        os.makedirs(os.path.dirname(filename))

    with open(filename, "w") as file:
        yaml.dump(data, file)


def view(data):
    r"""
    Print data formatted as YAML.
    """
    print(yaml.dump(data, default_flow_style=False, default_style=""))



r"""
Copy using scp.

(c) Tom de Geus, 2021, MIT
"""
from .external import exec_cmd


def from_remote(host, source, dest, verbose=False):
    r"""
    Copy a file from a remote system. Uses ``scp -p``.
    """

    cmd = f"scp -p {host:s}:{source:s} {dest:s}"

    exec_cmd(cmd, verbose)


def to_remote(host, source, dest, verbose=False):
    r"""
    Copy a file to a remote system. Uses ``scp -p``.
    """

    cmd = f"scp -p {source:s} {host:s}:{dest:s}"

    exec_cmd(cmd, verbose)



import argparse
import os
import pathlib
import re
import subprocess
import sys

import numpy as np

from .. import version
from .. import yaml
from .defaults import f_dump


def _shelephant_dump_parser():
    """
    Return parser for :py:func:`shelephant_dump`.
    """

    desc = "Dump filenames to a YAML-file."
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument("-o", "--output", default=f_dump, help="Output YAML-file.")
    parser.add_argument("-a", "--append", action="store_true", help="Append existing file.")
    parser.add_argument(
        "-e", "--exclude", type=str, action="append", help="Exclude files matching this pattern."
    )
    parser.add_argument(
        "-E",
        "--exclude-extension",
        type=str,
        action="append",
        default=[],
        help='Exclude files based on their extension (e.g. ".bak").',
    )
    parser.add_argument("--fmt", type=str, help='Formatter of each line, e.g. ``"mycmd {}"``.')
    parser.add_argument(
        "-c",
        "--command",
        action="store_true",
        help="Interpret the input as a command (instead of as filenames).",
    )
    parser.add_argument("-k", "--keep", type=str, action="append", help="Select files using regex.")
    parser.add_argument("--cwd", type=str, help="Directory to run the command in.")
    parser.add_argument(
        "--abspath",
        action="store_true",
        help="Store absolute paths (default: relative to the output file).",
    )
    parser.add_argument("-s", "--sort", action="store_true", help="Sort filenames.")
    parser.add_argument(
        "-f", "--force", action="store_true", help="Overwrite output file without prompt."
    )
    parser.add_argument("-v", "--version", action="version", version=version, help="")
    parser.add_argument("file", nargs="+", help="Files to list.")
    return parser


def shelephant_dump(args: list[str]):
    """
    Command-line tool, see ``--help``.
    :param args: Command-line arguments (should be all strings).
    """

    parser = _shelephant_dump_parser()
    args = parser.parse_args(args)

    prefix = os.path.dirname(args.output)
    files = args.file

    if args.command:
        cmd = " ".join(files)
        files = subprocess.check_output(cmd, shell=True, cwd=args.cwd).decode("utf-8").split("\n")
        files = list(filter(None, files))
        if args.cwd is not None:
            files = [os.path.join(args.cwd, file) for file in files]

    if args.abspath:
        files = [os.path.abspath(file) for file in files]
    else:
        files = [os.path.relpath(file, prefix) for file in files]

    if args.keep:
        ret = []
        for pattern in args.keep:
            ret += [file for file in files if re.match(pattern, file)]
        files = ret

    if args.exclude:
        excl = np.zeros(len(files), dtype=bool)
        for pattern in args.exclude:
            excl = np.logical_or(excl, np.array([re.match(pattern, file) for file in files]))
        files = [file for file, ex in zip(files, excl) if not ex]

    if args.exclude_extension:
        files = [file for file in files if pathlib.Path(file).suffix not in args.exclude_extension]

    if args.sort:
        files = sorted(files)

    if args.fmt:
        files = [args.fmt.format(file) for file in files]

    if args.append:
        main = yaml.read(args.output)
        assert type(main) == list, 'Can only append a "flat" file'
        files = main + files
        args.force = True

    yaml.dump(args.output, files, args.force)


def main():
    shelephant_dump(sys.argv[1:])


if __name__ == "__main__":
    main()



"""
Run as::

    python ssh_test.py HOST PREFIX
"""
import os
import subprocess
import sys
import unittest
from functools import partialmethod

from tqdm import tqdm

import shelephant

tqdm.__init__ = partialmethod(tqdm.__init__, disable=True)


def run(cmd):
    print(cmd)
    return subprocess.check_output(cmd, shell=True).decode("utf-8")


class Test_ssh(unittest.TestCase):
    def test_all(self):
        operations = [
            "bar.txt -> bar.txt",
            "foo.txt == foo.txt",
        ]

        output = run(
            (
                "shelephant_hostinfo -o myssh_send/shelephant_hostinfo.yaml --force "
                '--host "{:s}" --prefix "{:s}" -f -c'
            ).format(self.HOST, os.path.join(self.PREFIX, "myssh_get"))
        )

        output = run(
            "shelephant_send --detail --colors none --force "
            "myssh_send/shelephant_dump.yaml myssh_send/shelephant_hostinfo.yaml"
        )

        output = list(filter(None, output.split("\n")))

        self.assertEqual(output, operations)

        operations = [
            "bar.txt == bar.txt",
            "foo.txt -> foo.txt",
        ]

        output = run(
            (
                "shelephant_hostinfo -o myssh_send/shelephant_hostinfo.yaml --force "
                '--host "{:s}" --prefix "{:s}" -f'
            ).format(self.HOST, os.path.join(self.PREFIX, "myssh_get"))
        )

        output = run(
            "shelephant_send --detail --colors none --force "
            "myssh_send/shelephant_dump.yaml myssh_send/shelephant_hostinfo.yaml"
        )

        output = list(filter(None, output.split("\n")))
        self.assertEqual(output, operations)

        operations = {
            "<-": [],
            "->": [],
            "!=": [],
            "==": ["bar.txt", "foo.txt"],
        }

        output = run(
            "shelephant_diff -f --yaml shelephant_diff.yaml "
            "myssh_send/shelephant_dump.yaml myssh_send/shelephant_hostinfo.yaml"
        )

        output = shelephant.yaml.read("shelephant_diff.yaml")

        for key in operations:
            self.assertListEqual(output[key], operations[key])

        operations = [
            "bar.txt -> bar.txt",
            "foo.txt == foo.txt",
        ]

        output = run(
            (
                "shelephant_hostinfo -o myssh_get/shelephant_hostinfo.yaml --force "
                '--host "{:s}" --prefix "{:s}" -f -c'
            ).format(self.HOST, os.path.join(self.PREFIX, "myssh_send"))
        )

        output = run(
            "shelephant_get --detail --colors none --force myssh_get/shelephant_hostinfo.yaml"
        )

        output = list(filter(None, output.split("\n")))
        self.assertEqual(output, operations)

        os.remove("myssh_get/bar.txt")


if __name__ == "__main__":
    Test_ssh.PREFIX = sys.argv.pop()
    Test_ssh.HOST = sys.argv.pop()

    unittest.main()
